//+------------------------------------------------------------------+
//|                                                          THA.mq4 |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: Online Analytical Processing of trade hypercube |
//|                        https://www.mql5.com/en/blogs/post/685191 |
//+------------------------------------------------------------------+

#property copyright "Copyright  2016, Marketeer"
#property link      "https://www.mql5.com/en/blogs/post/685191"
#property version   "1.0"
#property description "Trade History Analysis: Non-trading Expert Adviser demonstrating OLAP classes"

#property strict

#include <OLAPcube.mqh>

enum SELECTORS
{
  SELECTOR_NONE,       // none
  SELECTOR_TYPE,       // type
  SELECTOR_SYMBOL,     // symbol
  SELECTOR_SERIAL,     // ordinal
  SELECTOR_MAGIC,      // magic
  SELECTOR_PROFITABLE, // profitable
  /* custom selector (see below) */
  SELECTOR_DURATION,   // duration in days
  /* all the next require a field as parameter */
  SELECTOR_WEEKDAY,    // day-of-week(datetime field)
  SELECTOR_DAYHOUR,    // hour-of-day(datetime field)
  SELECTOR_HOURMINUTE, // minute-of-hour(datetime field)
  SELECTOR_SCALAR      // scalar(field)
};

enum AGGREGATORS
{
  AGGREGATOR_SUM,      // SUM
  AGGREGATOR_AVERAGE,  // AVERAGE
  AGGREGATOR_MAX,      // MAX
  AGGREGATOR_MIN,      // MIN
  AGGREGATOR_COUNT     // COUNT
};

sinput string X = " X axis ";
input SELECTORS SelectorX = SELECTOR_SYMBOL;
input TRADE_RECORD_FIELDS FieldX = FIELD_NONE /* field does matter only for some selectors */;

sinput string Y = " Y axis ";
input SELECTORS SelectorY = SELECTOR_NONE;
input TRADE_RECORD_FIELDS FieldY = FIELD_NONE;

sinput string Z = " Z axis ";
input SELECTORS SelectorZ = SELECTOR_NONE;
input TRADE_RECORD_FIELDS FieldZ = FIELD_NONE;

sinput string F = " Filter ";
input SELECTORS Filter1 = SELECTOR_NONE;
input TRADE_RECORD_FIELDS Filter1Field = FIELD_NONE;
input float Filter1value1 = 0;
input float Filter1value2 = 0;

sinput string A = " Aggregator ";
input AGGREGATORS AggregatorType = AGGREGATOR_SUM;
input TRADE_RECORD_FIELDS AggregatorField = FIELD_PROFIT_AMOUNT;


int selectorCount;
SELECTORS selectorArray[4];
TRADE_RECORD_FIELDS selectorField[4];

#define PRT(A) Print(#A, " ", (A))

int OnInit()
{
  selectorCount = (SelectorX != SELECTOR_NONE) + (SelectorY != SELECTOR_NONE) + (SelectorZ != SELECTOR_NONE);
  PRT(selectorCount);
  selectorArray[0] = SelectorX;
  selectorArray[1] = SelectorY;
  selectorArray[2] = SelectorZ;
  selectorArray[3] = Filter1;
  selectorField[0] = FieldX;
  selectorField[1] = FieldY;
  selectorField[2] = FieldZ;
  selectorField[3] = Filter1Field;

  EventSetTimer(1);
  return(INIT_SUCCEEDED);
}

void OnTimer()
{
  process();
  EventKillTimer();
}

class DaysRangeSelector: public DateTimeSelector<TRADE_RECORD_FIELDS>
{
  protected:
    int granulatity;
    
  public:
    DaysRangeSelector(const int n): DateTimeSelector<TRADE_RECORD_FIELDS>(FIELD_DURATION, 7), granulatity(n)
    {
      _typename = typename(this);
    }
    
    virtual int getRange() const
    {
      return granulatity;
    }
    
    virtual bool select(const Record *r, int &index) const
    {
      double d = r.get(selector);
      int days = (int)(d / (60 * 60 * 24));
      index = MathMin(days, granulatity - 1);
      return true;
    }
    
    virtual string getLabel(const int index) const
    {
      return index < granulatity - 1 ? ((string)index + "D") : ((string)index + "D+");
    }
};

Selector<TRADE_RECORD_FIELDS> *createSelector(int i)
{
  switch(selectorArray[i])
  {
    case SELECTOR_TYPE:
      return new TypeSelector();
    case SELECTOR_SYMBOL:
      return new SymbolSelector();
    case SELECTOR_SERIAL:
      return new SerialNumberSelector();
    case SELECTOR_MAGIC:
      return new MagicSelector();
    case SELECTOR_PROFITABLE:
      return new ProfitableSelector();
    case SELECTOR_DURATION:
      return new DaysRangeSelector(15); // up to 14 days
    case SELECTOR_WEEKDAY:
      return new WeekDaySelector(selectorField[i]);
    case SELECTOR_DAYHOUR:
      return new DayHourSelector(selectorField[i]);
    case SELECTOR_HOURMINUTE:
      return new DayHourSelector(selectorField[i]);
    case SELECTOR_SCALAR:
      return new TradeSelector(selectorField[i]);
  }
  return NULL;
}

void process()
{
  HistoryDataAdapter history;
  Analyst<TRADE_RECORD_FIELDS> *analyst;
  
  Selector<TRADE_RECORD_FIELDS> *selectors[];
  ArrayResize(selectors, selectorCount);
  
  for(int i = 0; i < selectorCount; i++)
  {
    selectors[i] = createSelector(i);
  }
  Filter<TRADE_RECORD_FIELDS> *filters[];
  if(Filter1 != SELECTOR_NONE)
  {
    ArrayResize(filters, 1);
    Selector<TRADE_RECORD_FIELDS> *filterSelector = createSelector(3);
    if(Filter1value1 != Filter1value2)
    {
      filters[0] = new FilterRange<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1, Filter1value2);
    }
    else
    {
      filters[0] = new Filter<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1);
    }
  }
  
  Aggregator<TRADE_RECORD_FIELDS> *aggregator;
  
  // MQL does not support a 'class info' metaclass.
  // Otherwise we could use an array of classes instead of the switch
  switch(AggregatorType)
  {
    case AGGREGATOR_SUM:
      aggregator = new SumAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_AVERAGE:
      aggregator = new AverageAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_MAX:
      aggregator = new MaxAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_MIN:
      aggregator = new MinAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_COUNT:
      aggregator = new CountAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
  }
  
  LogDisplay display;
  
  analyst = new Analyst<TRADE_RECORD_FIELDS>(history, aggregator, display);
  
  analyst.acquireData();
  
  Print("Symbol number: ", TradeRecord::getSymbolCount());
  for(int i = 0; i < TradeRecord::getSymbolCount(); i++)
  {
    Print(i, "] ", TradeRecord::getSymbol(i));
  }

  Print("Magic number: ", TradeRecord::getMagicCount());
  for(int i = 0; i < TradeRecord::getMagicCount(); i++)
  {
    Print(i, "] ", TradeRecord::getMagic(i));
  }
  
  analyst.build();
  analyst.display();
  
  delete analyst;
  delete aggregator;
  for(int i = 0; i < selectorCount; i++)
  {
    delete selectors[i];
  }
  for(int i = 0; i < ArraySize(filters); i++)
  {
    delete filters[i].getSelector();
    delete filters[i];
  }
}

void OnTick()
{
}
